package de.hdm_stuttgart.mi.sda1.hashprovider;

import javax.crypto.SecretKey;
import javax.crypto.SecretKeyFactory;
import javax.crypto.spec.PBEKeySpec;

import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.security.spec.InvalidKeySpecException;

import org.apache.commons.codec.binary.Base64;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

/**
 * Credits to http://stackoverflow.com/users/90998/martin-konicek, answer #12 in
 * http://stackoverflow.com/questions/2860943/suggestions-for-library-to-hash-passwords-in-java
 * 
 * Subsequent code slightly modified from original.
 * 
 */
public class HashProvider {
   
   static private final Logger log = LogManager.getLogger(HashProvider.class);
   
    // The higher the number of iterations the more 
    // expensive computing the hash is for us
    // and also for a brute force attack.
    private static final int iterations = 10*1024;
    private static final int saltLen = 32;
    private static final int desiredKeyLen = 256;

    /**
     * Computes a salted PBKDF2 hash of given plain text password
        suitable for storing in a database
        
     * @param password The plain text password
     * @return the generated hash separated from salt by "$"
     */
    public static String getSaltedHash(char [] password) {
      try {
         final byte[] salt = SecureRandom.getInstance("SHA1PRNG").generateSeed(saltLen);
         // store the salt with the password
         return Base64.encodeBase64String(salt) + "$" + hash(password, salt);
      } catch (final NoSuchAlgorithmException e) {
         log.fatal("Unable to create hash: " + e);
      }
      return null;
    }

    /**
     * Checks whether given plain text password corresponds 
        to a stored salted hash of the password.
     * @param password The plain text password
     * @param stored the salt/hash to be verified against
     * @return true if input matches, false otherwise
     */
    public static boolean check(char[] password, String stored){
        final String[] saltAndPass = stored.split("\\$");
        if (saltAndPass.length != 2) {
           log.error("No salt detected!");
           return false;
        } else {
           final String hashOfInput = hash(password, Base64.decodeBase64(saltAndPass[0]));
           return null != hashOfInput && hashOfInput.equals(saltAndPass[1]);
        }
    }

    // using PBKDF2 from Sun, an alternative is https://github.com/wg/scrypt
    // cf. http://www.unlimitednovelty.com/2012/03/dont-use-bcrypt.html
    private static String hash(char [] password, byte[] salt) {
      try {
         final SecretKey key = 
               SecretKeyFactory.getInstance("PBKDF2WithHmacSHA1").generateSecret(
               new PBEKeySpec(password, salt, iterations, desiredKeyLen));
         
         return Base64.encodeBase64String(key.getEncoded());
      } catch (NoSuchAlgorithmException e) {
         log.fatal("Unable to create hash: " + e);
      } catch (InvalidKeySpecException e) {
         log.fatal("Unable to create hash: " + e);
      }
      return null;
    }
}