package de.hdm_stuttgart.mi.javastreams;

import java.util.List;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.function.Predicate;

/**
 * Loosely adapted and modified from
 * https://docs.oracle.com/javase/tutorial/java/javaOO/examples/Person.java
 * and
 * https://docs.oracle.com/javase/tutorial/java/javaOO/examples/RosterTest.java
 *
 */
public class RosterTest {
   
   static final List<Student> roster = Student.createRoster();
   
   interface CheckStudent {
       boolean test(Student p);
   }

   // Approach 1: Create Methods that Search for Persons that Match One
   // Characteristic

   public static void printPersonsOlderThan(List<Student> roster, int mark) {
       for (Student p : roster) {
           if (p.getMark()>= mark) {
               p.print();;
           }
       }
   }

   // Approach 2: Create More Generalized Search Methods

   public static void printPersonsWithinMarkRange(
       List<Student> roster, int low, int high) {
       for (Student p : roster) {
           if (low <= p.getMark() && p.getMark() <= high) {
               p.print();
           }
       }
   }

   // Approach 3: Specify Search Criteria Code in a Local Class
   // Approach 4: Specify Search Criteria Code in an Anonymous Class
   // Approach 5: Specify Search Criteria Code with a Lambda Expression

   public static void printPersons(
       List<Student> roster, CheckStudent tester) {
       for (Student p : roster) {
           if (tester.test(p)) {
               p.print();
           }
       }
   }

   // Approach 6: Use Standard Functional Interfaces with Lambda Expressions

   public static void printPersonsWithPredicate(
       List<Student> roster, Predicate<Student> tester) {
       for (Student p : roster) {
           if (tester.test(p)) {
               p.print();
           }
       }
   }

   // Approach 7: Use Lambda Expressions Throughout Your Application

   public static void processPersons(
       List<Student> roster,
       Predicate<Student> tester,
       Consumer<Student> block) {
       for (Student p : roster) {
           if (tester.test(p)) {
               block.accept(p);
           }
       }
   }

   // Approach 7, second example

   public static void processPersonsWithFunction(
       List<Student> roster,
       Predicate<Student> tester,
       Function<Student, String> mapper,
       Consumer<String> block) {
       for (Student p : roster) {
           if (tester.test(p)) {
               String data = mapper.apply(p);
               block.accept(data);
           }
       }
   }
    
   // Approach 8: Use Generics More Extensively

   public static <X, Y> void processElements(
       Iterable<X> source,
       Predicate<X> tester,
       Function<X, Y> mapper,
       Consumer<Y> block) {
           for (X p : source) {
               if (tester.test(p)) {
                   Y data = mapper.apply(p);
                   block.accept(data);
               }
           }
   }

   public static void main(String... args) {

       for (Student p : roster) {
           p.print();
       }

       // Approach 1: Create Methods that Search for Persons that Match One
       // Characteristic

       System.out.println("Persons older than 20:");
       printPersonsOlderThan(roster, 20);
       System.out.println();

       // Approach 2: Create More Generalized Search Methods

       System.out.println("Persons between the ages of 14 and 30:");
       printPersonsWithinMarkRange(roster, 14, 30);
       System.out.println();

       // Approach 3: Specify Search Criteria Code in a Local Class

       System.out.println("Persons who are eligible for Selective Service:");

       class CheckStudentEligibleForSelectiveService implements CheckStudent {
          public boolean test(Student p) {
               return p.getGender() == Student.Sex.MALE
                   && p.getMark() >= 2
                   && p.getMark() <= 3;
           }
       }

       printPersons(
           roster, new CheckStudentEligibleForSelectiveService());


       System.out.println();

       // Approach 4: Specify Search Criteria Code in an Anonymous Class

       System.out.println("Persons who are eligible for Selective Service " +
           "(anonymous class):");

       printPersons(
           roster,
           new CheckStudent() {
               public boolean test(Student p) {
                   return p.getGender() == Student.Sex.MALE
                       && p.getMark() >= 2
                       && p.getMark() <= 3;
               }
           }
       );

       System.out.println();

       // Approach 5: Specify Search Criteria Code with a Lambda Expression

       System.out.println("Persons who are eligible for Selective Service " +
           "(lambda expression):");

       printPersons(
           roster,
           (Student p) -> p.getGender() == Student.Sex.MALE
               && p.getMark() >= 2
               && p.getMark() <= 3
       );

       System.out.println();

       // Approach 6: Use Standard Functional Interfaces with Lambda
       // Expressions

       System.out.println("Persons who are eligible for Selective Service " +
           "(with Predicate parameter):");

       printPersonsWithPredicate(
           roster,
           p -> p.getGender() == Student.Sex.MALE
               && p.getMark() >= 18
               && p.getMark() <= 25
       );

       System.out.println();

       // Approach 7: Use Lamba Expressions Throughout Your Application

       System.out.println("Persons who are eligible for Selective Service " +
           "(with Predicate and Consumer parameters):");

       processPersons(
           roster,
           p -> p.getGender() == Student.Sex.MALE
               && p.getMark() >= 2
               && p.getMark() <= 3,
           p -> p.print()
       );

       System.out.println();

       // Approach 7, second example

       System.out.println("Persons who are eligible for Selective Service " +
           "(with Predicate, Function, and Consumer parameters):");

       processPersonsWithFunction(
           roster,
           p -> p.getGender() == Student.Sex.MALE
               && p.getMark() >= 2
               && p.getMark() <= 3,
           p -> p.getEmailAddress(),
           email -> System.out.println(email)
       );

       System.out.println();

       // Approach 8: Use Generics More Extensively

       System.out.println("Persons who are eligible for Selective Service " +
           "(generic version):");

       processElements(
           roster,
           p -> p.getGender() == Student.Sex.MALE
               && p.getMark() >= 2
               && p.getMark() <= 3,
           p -> p.getEmailAddress(),
           email -> System.out.println(email)
       );

       System.out.println();

       // Approach 9: Use Bulk Data Operations That Accept Lambda Expressions
       // as Parameters

       System.out.println("Persons who are eligible for Selective Service " +
           "(with bulk data operations):");

       roster
           .stream()
           .filter(
               p -> p.getGender() == Student.Sex.MALE
                   && p.getMark() >= 2
                   && p.getMark() <= 3)
           .map(p -> p.getEmailAddress())
           .forEach(email -> System.out.println(email));
    }
}
